#include "error.h"
#include "gl.h"
#include "image8.h"
#include "shader.h"
#include <gl/gl.h>
#define GSA_NOMAIN
#include "gsa.h"

struct gsa_sprite sprites[256];

struct sw_image8 *_gfx;
struct sw_window *_win;

static char *vert_source = "#version 450\n\
in vec2 pos;\
in vec2 tex_coord;\
layout(location = 0) out vec2 out_tex_coord;\
void main() {\
gl_Position = vec4((pos / vec2(152, -88)) - vec2(1, -1), 0.f, 1.f);\
out_tex_coord = tex_coord;\
}\
";

static char *frag_source = "#version 450\n\
layout(location = 0) in vec2 tex_coord;\
uniform sampler2D tex;\
out vec4 color;\
void main(){\
float r = texture(tex, tex_coord).r;\
if(r > 0)\
color = vec4(0.9, 0.2, 0.1, 1.0);\
}\
";

static struct sw_shaderprogram program;

static u32 vbo, vao;

#define MAX_RENDER_VERTS 60

struct render_vert {
	f32 x, y, tx, ty;
};

static i32 next_render_vert;
static struct render_vert render_verts[MAX_RENDER_VERTS * 4];

int gsa_main(int argc, char *argv[]) {
	(void)argc;
	(void)argv;
	u32 tex;
	struct sw_image8 *img;

	sw_log("Initialising GameSkunkAdvance v0.0");

	_win = sw_window_create(sw_vec2i(304, 176), "Game Skunk Advance v0.0");
	program = sw_shaderprogram_create(vert_source, frag_source);
	sw_shaderprogram_use(program);

	img = sw_image8_load_png("gfx.png");
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(
	    GL_TEXTURE_2D,
	    0,
	    GL_RED,
	    4096,
	    4096,
	    0,
	    GL_RED,
	    GL_UNSIGNED_BYTE,
	    img->_data
	);
	sw_image8_destroy(img);

	glGenBuffers(1, &vao);
	glBindVertexArray(vao);

	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexAttribPointer(
	    glGetAttribLocation(program._program, "pos"),
	    2,
	    GL_FLOAT,
	    GL_FALSE,
	    sizeof(struct render_vert),
	    0
	);
	glVertexAttribPointer(
	    glGetAttribLocation(program._program, "tex_coord"),
	    2,
	    GL_FLOAT,
	    GL_FALSE,
	    sizeof(struct render_vert),
	    (void *)(2 * sizeof(float))
	);
	glEnableVertexAttribArray(glGetAttribLocation(program._program, "pos"));
	glEnableVertexAttribArray(glGetAttribLocation(program._program, "tex_coord")
	);

	init();
	sw_window_run(_win, _gsa_tick);

	return 0;
}

static void rect(f32 x, f32 y, f32 w, f32 h, i16 tile) {
	f32 tx, ty, ts;

	tx = ((tile % 256) * 16.f) / 4096.f;
	ty = ((tile / 256) * 16.f) / 4096.f;
	ts = 1.f / 256.f;

	if(next_render_vert >= MAX_RENDER_VERTS) {
		sw_error(
		    "trying to draw more than the current max of %i vertices in a "
		    "frame",
		    MAX_RENDER_VERTS
		);
	}

	sw_log("tx %f | ty %f | ts %f", tx, ty, ts);

	render_verts[next_render_vert].x = x;
	render_verts[next_render_vert].y = y;
	render_verts[next_render_vert].tx = tx;
	render_verts[next_render_vert].ty = ty;
	++next_render_vert;

	render_verts[next_render_vert].x = x + w;
	render_verts[next_render_vert].y = y;
	render_verts[next_render_vert].tx = tx + ts;
	render_verts[next_render_vert].ty = ty;
	++next_render_vert;

	render_verts[next_render_vert].x = x;
	render_verts[next_render_vert].y = y + h;
	render_verts[next_render_vert].tx = tx;
	render_verts[next_render_vert].ty = ty + ts;
	++next_render_vert;

	render_verts[next_render_vert].x = x + w;
	render_verts[next_render_vert].y = y;
	render_verts[next_render_vert].tx = tx + ts;
	render_verts[next_render_vert].ty = ty;
	++next_render_vert;

	render_verts[next_render_vert].x = x;
	render_verts[next_render_vert].y = y + h;
	render_verts[next_render_vert].tx = tx;
	render_verts[next_render_vert].ty = ty + ts;
	++next_render_vert;

	render_verts[next_render_vert].x = x + w;
	render_verts[next_render_vert].y = y + h;
	render_verts[next_render_vert].tx = tx + ts;
	render_verts[next_render_vert].ty = ty + ts;
	++next_render_vert;
}

void _gsa_tick() {
	i32 i;

	tick();

	next_render_vert = 0;
	/* glBegin(GL_TRIANGLES); */
	for(i = 0; i < MAX_SPRITES; ++i) {
		if(sprites[i].tile > 0) {
			rect(sprites[i].x, sprites[i].y, 16, 16, sprites[i].tile);
		}
	}

	glBindVertexArray(vao);
	glBufferData(
	    GL_ARRAY_BUFFER,
	    sizeof(struct render_vert) * (next_render_vert),
	    render_verts,
	    GL_DYNAMIC_DRAW
	);
	glDrawArrays(GL_TRIANGLES, 0, next_render_vert);
	/* glEnd(); */
}
